/*
 * \brief  Azure SDK IoT Hub client
 * \author Timo Nicolai
 * \date   2023-02-23
 */

/*
 * Copyright (C) 2023 gapfruit AG
 */

#pragma once

/* Azure includes */
extern "C" {

#include <azure/az_iot.h>

}

/* local includes */
#include <azure_iot_client/azure/check.h>
#include <azure_iot_client/azure/types.h>
#include <azure_iot_client/azure/util.h>
#include <azure_iot_client/iot_client.h>
#include <azure_iot_client/mqtt/types.h>

namespace Azure {
	class IoT_hub_client;
}


class Azure::IoT_hub_client : protected IoT_client<::az_iot_hub_client>
{
	public:

		using Direct_method_request = ::az_iot_hub_client_method_request;
		using Twin_response         = ::az_iot_hub_client_twin_response;

		static constexpr char const *FILE_UPLOAD_API_VERSION = "2020-03-13";

	private:

		IoT_hub_hostname _iot_hub_hostname;
		Device_id        _device_id;

		Genode::size_t _mqtt_twin_document_topic_request_id { 0 };
		Genode::size_t _mqtt_twin_patch_topic_request_id    { 0 };

	public:

		explicit IoT_hub_client(IoT_hub_hostname const &iot_hub_hostname,
		                        Device_id const &device_id)
		:
			_iot_hub_hostname { iot_hub_hostname },
			_device_id { device_id }
		{
			check(::az_iot_hub_client_init(&_client,
			                               Util::to_span(_iot_hub_hostname),
			                               Util::to_span(_device_id),
			                               nullptr),
			      "initialize IoT Hub client");
		}

		MQTT::Client_id mqtt_client_id() const
		{
			return _get<MQTT::Client_id>("get IoT Hub MQTT client id",
			                             ::az_iot_hub_client_get_client_id);
		}

		MQTT::Hostname mqtt_hostname() const
		{
			return _iot_hub_hostname;
		}

		MQTT::Username mqtt_username() const
		{
			return _get<MQTT::Username>("get IoT Hub MQTT username",
			                            ::az_iot_hub_client_get_user_name);
		}

		Direct_method_request mqtt_parse_direct_method_request(
			MQTT::Topic const &topic) const
		{
			::az_iot_hub_client_method_request request;
			check(::az_iot_hub_client_methods_parse_received_topic(&_client,
			                                                       Util::to_span(topic),
			                                                       &request),
			      "parse direct method request");

			return request;
		}

		MQTT::Topic mqtt_direct_method_response_topic(
			Direct_method_request const &request,
			Genode::uint16_t status) const
		{
			return _get<MQTT::Topic>(
				"get direct method response topic",
				::az_iot_hub_client_methods_response_get_publish_topic,
				request.request_id,
				status);
		}

		MQTT::Topic mqtt_twin_document_topic()
		{
			MQTT::Request_id request_id {
				"twin_document_",
				++_mqtt_twin_document_topic_request_id };

			return _get<MQTT::Topic>(
				"get twin document topic",
				::az_iot_hub_client_twin_document_get_publish_topic,
				Util::to_span(request_id));
		}

		MQTT::Topic mqtt_twin_patch_topic()
		{
			MQTT::Request_id request_id {
				"twin_patch_",
				++_mqtt_twin_patch_topic_request_id };

			return _get<MQTT::Topic>(
				"get twin patch topic",
				::az_iot_hub_client_twin_patch_get_publish_topic,
				Util::to_span(request_id));
		}

		Twin_response mqtt_parse_twin_response(MQTT::Topic const &topic) const
		{
			::az_iot_hub_client_twin_response response;
			check(::az_iot_hub_client_twin_parse_received_topic(&_client,
			                                                    Util::to_span(topic),
			                                                    &response),
			      "parse twin response");

			return response;
		}

		Url http_file_upload_endpoint() const
		{
			return Url {
				"https://",
				_iot_hub_hostname,
				"/devices/",
				_device_id,
				"/files?api-version=",
				FILE_UPLOAD_API_VERSION
			};
		}

		Url http_file_upload_notify_endpoint() const
		{
			return Url {
				"https://",
				_iot_hub_hostname,
				"/devices/",
				_device_id,
				"/files/notifications/?api-version=",
				FILE_UPLOAD_API_VERSION
			};
		}
};
