/*
 * \brief  Wrapper around `pthread_cond_t`
 * \author Timo Nicolai
 * \date   2023-03-05
 */

/*
 * Copyright (C) 2023 gapfruit AG
 */

#pragma once

/* Genode includes */
#include <base/exception.h>
#include <base/log.h>

/* libc includes */
extern "C" {

#include <pthread.h>

}

/* local includes */
#include <azure_iot_client/util/mutex.h>

namespace Util {
	class Condition_variable;
}


class Util::Condition_variable
{
	private:

		::pthread_cond_t  _cond     { };
		Mutex             _mutex    { };
		bool              _signaled { false };

		void _pthread_check(int rc, char const *what)
		{
			if (rc != 0) {
				Genode::error("pthread: failed to ", what);
				throw Exception { };
			}
		}

	public:

		struct Exception : Genode::Exception { };

		explicit Condition_variable()
		{
			_pthread_check(::pthread_cond_init(&_cond, nullptr),
			               "initialize condition variable");
		}

		~Condition_variable()
		{
			::pthread_cond_destroy(&_cond);
		}

		Condition_variable(Condition_variable const &)    = delete;
		Condition_variable &operator=(Condition_variable) = delete;

		void wait()
		{
			Mutex::Guard guard { _mutex };

			while (!_signaled) {
				_pthread_check(::pthread_cond_wait(&_cond, &_mutex._mutex),
				               "wait on condition variable");
			}
		}

		void signal()
		{
			{
				Mutex::Guard guard { _mutex };

				_signaled = true;
			}

			_pthread_check(::pthread_cond_signal(&_cond),
			               "signal condition variable");
		}
};
