/*
 * \brief  Wrapper around `pthread_mutex_t`
 * \author Timo Nicolai
 * \date   2023-03-05
 */

/*
 * Copyright (C) 2023 gapfruit AG
 */

#pragma once

/* Genode includes */
#include <base/exception.h>
#include <base/log.h>

/* libc includes */
extern "C" {

#include <pthread.h>

}

namespace Util {
	class Mutex;
}


class Util::Mutex
{
	friend class Condition_variable;

	private:

		::pthread_mutex_t _mutex { };

		void _pthread_check(int rc, char const *what)
		{
			if (rc != 0) {
				Genode::error("pthread: failed to ", what);
				throw Exception { };
			}
		}

	public:

		struct Exception : Genode::Exception { };

		class Guard
		{
			private:

				Mutex &_mutex;

			public:

				explicit Guard(Mutex &mutex)
				:
					_mutex { mutex }
				{
					_mutex.lock();
				}

				~Guard()
				{
					_mutex.unlock();
				}
		};

		explicit Mutex()
		{
			_pthread_check(::pthread_mutex_init(&_mutex, nullptr),
			               "initialize mutex");
		}

		~Mutex()
		{
			::pthread_mutex_destroy(&_mutex);
		}

		Mutex(Mutex const &)    = delete;
		Mutex &operator=(Mutex) = delete;

		void lock()
		{
			_pthread_check(::pthread_mutex_lock(&_mutex),
			               "lock mutex");
		}

		void unlock()
		{
			_pthread_check(::pthread_mutex_unlock(&_mutex),
			               "unlock mutex");
		}
};
